@@ -1,4 +1,4 @@
-# Copyright (c) 2018, Neil Booth
+# Copyright (c) 2018-2019, Neil Booth
 #
 # All rights reserved.
 #
@@ -24,19 +24,18 @@
 # WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
 
-__all__ = ('Connector', 'RPCSession', 'MessageSession', 'Server', 'ExcessiveSessionCostError',
-           'BatchError', 'Concurrency', 'ReplyAndDisconnect')
+__all__ = ('RPCSession', 'MessageSession', 'ExcessiveSessionCostError',
+           'BatchError', 'Concurrency', 'ReplyAndDisconnect', 'SessionKind')
 
 
 import asyncio
-from contextlib import suppress
+from enum import Enum
 import logging
 from math import ceil
 import time
 
 from aiorpcx.curio import (
-    Event, TaskGroup, TaskTimeout, CancelledError,
-    timeout_after, spawn_sync, ignore_after, sleep
+    TaskGroup, TaskTimeout, CancelledError, timeout_after, sleep
 )
 from aiorpcx.framing import (
     NewlineFramer, BitcoinFramer, BadMagicError, BadChecksumError, OversizedPayloadError
@@ -45,50 +44,18 @@ from aiorpcx.jsonrpc import (
     Request, Batch, Notification, ProtocolError, RPCError,
     JSONRPC, JSONRPCv2, JSONRPCConnection
 )
-from aiorpcx.util import NetAddress
 
 
 class ReplyAndDisconnect(Exception):
     '''Force a session disconnect after sending result (a Python object or an RPCError).
     '''
 
-    def __init__(self, result):
-        self.result = result
-
-
-class Connector(object):
-
-    def __init__(self, session_factory, host=None, port=None, proxy=None,
-                 **kwargs):
-        self.session_factory = session_factory
-        self.host = host
-        self.port = port
-        self.proxy = proxy
-        self.protocol = None
-        self.loop = kwargs.get('loop', asyncio.get_event_loop())
-        self.kwargs = kwargs
-        # By default, do not limit outgoing connections
-        self.cost_hard_limit = 0
-
-    async def create_connection(self):
-        '''Initiate a connection.'''
-        connector = self.proxy or self.loop
-        return await connector.create_connection(
-            self.session_factory, self.host, self.port, **self.kwargs)
-
-    async def __aenter__(self):
-        _transport, self.protocol = await self.create_connection()
-        return self.protocol
-
-    async def __aexit__(self, exc_type, exc_value, traceback):
-        await self.protocol.close()
-
 
 class ExcessiveSessionCostError(RuntimeError):
     pass
 
 
-class Concurrency(object):
+class Concurrency:
 
     def __init__(self, target):
         self._target = int(target)
@@ -120,18 +87,16 @@ class Concurrency(object):
             self._semaphore.release()
 
 
-class SessionBase(asyncio.Protocol):
+class SessionKind(Enum):
+    CLIENT = 'client'
+    SERVER = 'server'
+
+
+class SessionBase:
     '''Base class of networking sessions.
 
     There is no client / server distinction other than who initiated
     the connection.
-
-    To initiate a connection to a remote server pass host, port and
-    proxy to the constructor, and then call create_connection().  Each
-    successful call should have a corresponding call to close().
-
-    Alternatively if used in a with statement, the connection is made
-    on entry to the block, and closed on exit from the block.
     '''
 
     # Multiply this by bandwidth bytes used to get resource usage cost
@@ -153,21 +118,12 @@ class SessionBase(asyncio.Protocol):
     # Force-close a connection if its socket send buffer stays full this long
     max_send_delay = 20.0
 
-    def __init__(self, *, framer=None, loop=None):
-        self.framer = framer or self.default_framer()
+    def __init__(self, transport, *, loop=None):
+        self.transport = transport
         self.loop = loop or asyncio.get_event_loop()
         self.logger = logging.getLogger(self.__class__.__name__)
-        self.transport = None
-        self.closed_event = self.event()
-        # Set when a connection is made
-        self._proxy = None
-        self._remote_address = None
         # For logger.debug messsages
         self.verbosity = 0
-        # Cleared when the send socket is full
-        self._can_send = self.event()
-        self._can_send.set()
-        self._task = None
         self._group = TaskGroup()
         # Statistics.  The RPC object also keeps its own statistics.
         self.start_time = time.time()
@@ -185,93 +141,42 @@ class SessionBase(asyncio.Protocol):
         self._cost_fraction = 0.0
         # Concurrency control for incoming request handling
         self._incoming_concurrency = Concurrency(self.initial_concurrent)
-
-    async def _process_messages(self):
-        async with self._group:
-            await self._group.spawn(self._receive_messages)
-
-    def _receive_messages(self):
-        raise NotImplementedError
-
-    async def _limited_wait(self, secs):
-        # Wait at most secs seconds to send, otherwise abort the connection
-        try:
-            async with timeout_after(secs):
-                await self._can_send.wait()
-        except TaskTimeout:
-            self.abort()
-            raise
+        # By default, do not limit outgoing connections
+        if self.session_kind == SessionKind.CLIENT:
+            self.cost_hard_limit = 0
 
     async def _send_message(self, message):
-        if not self._can_send.is_set():
-            await self._limited_wait(self.max_send_delay)
-        if not self.is_closing():
-            framed_message = self.framer.frame(message)
-            self.send_size += len(framed_message)
-            self.bump_cost(len(framed_message) * self.bw_cost_per_byte)
-            self.send_count += 1
-            self.last_send = time.time()
-            if self.verbosity >= 4:
-                self.logger.debug(f'Sending framed message {framed_message}')
-            self.transport.write(framed_message)
+        if self.verbosity >= 4:
+            self.logger.debug(f'sending message {message}')
+        try:
+            async with timeout_after(self.max_send_delay):
+                await self.transport.write(message)
+        except TaskTimeout:
+            await self.abort()
+            raise
+        self.send_size += len(message)
+        self.bump_cost(len(message) * self.bw_cost_per_byte)
+        self.send_count += 1
+        self.last_send = time.time()
         return self.last_send
 
     def _bump_errors(self, exception=None):
         self.errors += 1
         self.bump_cost(self.error_base_cost + getattr(exception, 'cost', 0.0))
 
-    # asyncio framework
-    def data_received(self, framed_message):
-        '''Called by asyncio when a message comes in.'''
-        if self.verbosity >= 4:
-            self.logger.debug(f'Received framed message {framed_message}')
-        self.recv_size += len(framed_message)
-        self.bump_cost(len(framed_message) * self.bw_cost_per_byte)
-        self.framer.received_bytes(framed_message)
+    @property
+    def session_kind(self):
+        '''Either client or server.'''
+        return self.transport.kind
 
-    def pause_writing(self):
-        '''Transport calls when the send buffer is full.'''
-        if not self.is_closing():
-            self._can_send.clear()
-            self.transport.pause_reading()
+    async def connection_lost(self):
+        pass
 
-    def resume_writing(self):
-        '''Transport calls when the send buffer has room.'''
-        if not self._can_send.is_set():
-            self._can_send.set()
-            self.transport.resume_reading()
-
-    def connection_made(self, transport):
-        '''Called by asyncio when a connection is established.
-
-        Derived classes overriding this method must call this first.'''
-        self.transport = transport
-        # If the Socks proxy was used then _proxy and _remote_address are already set
-        if self._proxy is None:
-            # This would throw if called on a closed SSL transport.  Fixed in asyncio in
-            # Python 3.6.1 and 3.5.4
-            peername = transport.get_extra_info('peername')
-            self._remote_address = NetAddress(peername[0], peername[1])
-        self._task = spawn_sync(self._process_messages(), loop=self.loop)
-
-    def connection_lost(self, exc):
-        '''Called by asyncio when the connection closes.
-
-        Tear down things done in connection_made.'''
-        # Work around uvloop bug; see https://github.com/MagicStack/uvloop/issues/246
-        if self.transport:
-            self.transport = None
-            self.closed_event.set()
-            # Release waiting tasks
-            self._can_send.set()
-            # Cancelling directly leads to self-cancellation problems for member
-            # functions await-ing self.close()
-            self.loop.call_soon(self._task.cancel)
-
-    # External API
-    def is_send_buffer_full(self):
-        '''Return True if the send socket buffer is full.'''
-        return not self._can_send.is_set()
+    def data_received(self, data):
+        if self.verbosity >= 2:
+            self.logger.debug(f'received data {data}')
+        self.recv_size += len(data)
+        self.bump_cost(len(data) * self.bw_cost_per_byte)
 
     def bump_cost(self, delta):
         # Delta can be positive or negative
@@ -304,6 +209,13 @@ class SessionBase(asyncio.Protocol):
             self.logger.info(f'changing task concurrency from {value} to {target}')
         self._incoming_concurrency.set_target(target)
 
+    async def _process_messages(self, recv_message):
+        raise NotImplementedError
+
+    async def process_messages(self, recv_message):
+        async with self._group:
+            await self._group.spawn(self._process_messages, recv_message)
+
     def unanswered_request_count(self):
         '''The number of requests received but not yet answered.'''
         # Max with zero in case the message processing task hasn't yet spawned
@@ -321,47 +233,33 @@ class SessionBase(asyncio.Protocol):
 
     def proxy(self):
         '''Returns the proxy used, or None.'''
-        return self._proxy
+        return self.transport.proxy()
 
     def remote_address(self):
         '''Returns a NetAddress or None if not connected.'''
-        return self._remote_address
+        return self.transport.remote_address()
 
     def is_closing(self):
         '''Return True if the connection is closing.'''
-        return not self.transport or self.transport.is_closing()
+        return self.transport.is_closing()
 
-    def event(self):
-        return Event(loop=self.loop)
-
-    def abort(self):
+    async def abort(self):
         '''Forcefully close the connection.'''
-        if self.transport:
-            self.transport.abort()
+        await self.transport.abort()
 
     async def close(self, *, force_after=30):
         '''Close the connection and return when closed.'''
-        if self.transport:
-            self.transport.close()
-            try:
-                async with timeout_after(force_after):
-                    await self.closed_event.wait()
-            except TaskTimeout:
-                self.abort()
-                await self.closed_event.wait()
+        await self.transport.close(force_after)
 
 
 class MessageSession(SessionBase):
     '''Session class for protocols where messages are not tied to responses,
     such as the Bitcoin protocol.
-
-    To use as a client (connection-opening) session, pass host, port
-    and perhaps a proxy.
     '''
-    async def _receive_messages(self):
-        while not self.is_closing():
+    async def _process_messages(self, recv_message):
+        while True:
             try:
-                message = await self.framer.receive_message()
+                message = await recv_message()
             except BadMagicError as e:
                 magic, expected = e.args
                 self.logger.error(
@@ -369,7 +267,8 @@ class MessageSession(SessionBase):
                     f'disconnecting'
                 )
                 self._bump_errors(e)
-                await self.close()
+                await self._group.spawn(self.close)
+                await sleep(0.001)
             except OversizedPayloadError as e:
                 command, payload_len = e.args
                 self.logger.error(
@@ -377,7 +276,8 @@ class MessageSession(SessionBase):
                     f'{command}, disconnecting'
                 )
                 self._bump_errors(e)
-                await self.close()
+                await self._group.spawn(self.close)
+                await sleep(0.001)
             except BadChecksumError as e:
                 payload_checksum, claimed_checksum = e.args
                 self.logger.warning(
@@ -413,10 +313,9 @@ class MessageSession(SessionBase):
             self.logger.exception(f'exception handling {message}')
             self._bump_errors()
 
-    # External API
     def default_framer(self):
         '''Return a bitcoin framer.'''
-        return BitcoinFramer(bytes.fromhex('e3e1f3e8'), 128_000_000)
+        return BitcoinFramer()
 
     async def handle_message(self, message):
         '''message is a (command, payload) pair.'''
@@ -433,7 +332,7 @@ class BatchError(Exception):
         self.request = request   # BatchRequest object
 
 
-class BatchRequest(object):
+class BatchRequest:
     '''Used to build a batch request to send to the server.  Stores
     the
 
@@ -488,8 +387,8 @@ class BatchRequest(object):
     async def __aexit__(self, exc_type, exc_value, traceback):
         if exc_type is None:
             self.batch = Batch(self._requests)
-            message, event = self._session.connection.send_batch(self.batch)
-            self.results = await self._session._send_concurrent(message, event, len(self.batch))
+            message, future = self._session.connection.send_batch(self.batch)
+            self.results = await self._session._send_concurrent(message, future, len(self.batch))
             if self._raise_errors:
                 if any(isinstance(item, Exception) for item in self.results):
                     raise BatchError(self)
@@ -507,8 +406,8 @@ class RPCSession(SessionBase):
     sent_request_timeout = 30.0
     log_me = False
 
-    def __init__(self, *, framer=None, loop=None, connection=None):
-        super().__init__(framer=framer, loop=loop)
+    def __init__(self, transport, *, loop=None, connection=None):
+        super().__init__(transport, loop=loop)
         self.connection = connection or self.default_connection()
         # Concurrency control for outgoing request sending
         self._outgoing_concurrency = Concurrency(50)
@@ -526,10 +425,10 @@ class RPCSession(SessionBase):
             self.logger.info(f'changing outgoing request concurrency to {target} from {current}')
             self._outgoing_concurrency.set_target(target)
 
-    async def _receive_messages(self):
-        while not self.is_closing():
+    async def _process_messages(self, recv_message):
+        while True:
             try:
-                message = await self.framer.receive_message()
+                message = await recv_message()
             except MemoryError as e:
                 self.logger.warning(f'{e!r}')
                 continue
@@ -568,7 +467,7 @@ class RPCSession(SessionBase):
             self.logger.info(f'incoming request {request} timed out after {timeout} secs')
             result = RPCError(JSONRPC.SERVER_BUSY, 'server busy - request timed out')
         except ReplyAndDisconnect as e:
-            result = e.result
+            result = e.args[0]
             disconnect = True
         except ExcessiveSessionCostError:
             result = RPCError(JSONRPC.EXCESSIVE_RESOURCE_USAGE, 'excessive resource usage')
@@ -588,12 +487,12 @@ class RPCSession(SessionBase):
         if disconnect:
             await self.close()
 
-    async def _send_concurrent(self, message, event, request_count):
+    async def _send_concurrent(self, message, future, request_count):
         async with self._outgoing_concurrency:
             send_time = await self._send_message(message)
             try:
                 async with timeout_after(self.sent_request_timeout):
-                    await event.wait()
+                    return await future
             finally:
                 time_taken = time.time() - send_time
                 if request_count == 1:
@@ -603,16 +502,10 @@ class RPCSession(SessionBase):
                 if len(self._req_times) >= self.recalibrate_count:
                     self._recalc_concurrency()
 
-        result = event.result
-        # For batches, CancelledError can happen with cancel_pending_requests
-        if isinstance(result, Exception):
-            raise result
-        return result
-
-    def connection_lost(self, exc):
+    async def connection_lost(self):
+        await super().connection_lost()
         # Cancel pending requests and message processing
         self.connection.cancel_pending_requests()
-        super().connection_lost(exc)
 
     # External API
     def default_connection(self):
@@ -628,8 +521,8 @@ class RPCSession(SessionBase):
 
     async def send_request(self, method, args=()):
         '''Send an RPC request over the network.'''
-        message, event = self.connection.send_request(Request(method, args))
-        return await self._send_concurrent(message, event, 1)
+        message, future = self.connection.send_request(Request(method, args))
+        return await self._send_concurrent(message, future, 1)
 
     async def send_notification(self, method, args=()):
         '''Send an RPC notification over the network.'''
@@ -651,29 +544,3 @@ class RPCSession(SessionBase):
         BatchRequest doc string.
         '''
         return BatchRequest(self, raise_errors)
-
-
-class Server(object):
-    '''A simple wrapper around an asyncio.Server object.'''
-
-    def __init__(self, session_factory, host=None, port=None, *,
-                 loop=None, **kwargs):
-        self.host = host
-        self.port = port
-        self.loop = loop or asyncio.get_event_loop()
-        self.server = None
-        self._session_factory = session_factory
-        self._kwargs = kwargs
-
-    async def listen(self):
-        self.server = await self.loop.create_server(
-            self._session_factory, self.host, self.port, **self._kwargs)
-
-    async def close(self):
-        '''Close the listening socket.  This does not close any ServerSession
-        objects created to handle incoming connections.
-        '''
-        if self.server:
-            self.server.close()
-            await self.server.wait_closed()
-            self.server = None
